/* -*- c++ -*- */
#ifndef MPS_SocketDescriptor_H
#define MPS_SocketDescriptor_H

#include <ref.h>
#include <exception>
#include <string>
#include <map>

#include <FileDescriptor.h>

class SocketDescriptor;
class ClientSocketDescriptor;
class ServerSocketDescriptor;

/**
 * Subclasses FileDescriptor to provide a general-purpose wrapper for
 * BSD TCP or UDP server- or client-sockets. */
class SocketDescriptor: public FileDescriptor {
private:
  string hostname_local;	///< Name of the IP-address of the local end of the socket
  int port_local;		///< Port number of the local end of the socket

  string hostname_remote;	///< Name of the IP-address of the remote end of the socket
  int port_remote;		///< Port number of the remote end of the socket

protected:
  /// Override to use ::recv and ::send if on linux or win32.
  virtual int lowlevel_read(int fd, char *buf, int len);
  virtual int lowlevel_write(int fd, char const *buf, int len);

#ifdef __WIN32__
  virtual int lowlevel_close(int fd);
#endif

  friend class ServerSocketDescriptor;
  SocketDescriptor(ref<FileDescriptorManager> const &_fdMgr, int _fd)
    : FileDescriptor(_fdMgr, _fd),
      hostname_local("UNKNOWN"),
      port_local(0),
      hostname_remote("UNKNOWN"),
      port_remote(0)
  {}

  void setLocalAddr(string const &hostname, int port) {
    hostname_local = hostname;
    port_local = port;
  }

  void setRemoteAddr(string const &hostname, int port) {
    hostname_remote = hostname;
    port_remote = port;
  }

public:
  virtual void close();		///< Overridden to deal with shutdown() before close()

  /** @name AddressAccessors
   *
   * Accessors for local and remote address information for this
   * socket. */
  //@{
  string const &getLocalHostname() const { return hostname_local; }
  int getLocalPort() const { return port_local; }
  string const &getRemoteHostname() const { return hostname_remote; }
  int getRemotePort() const { return port_remote; }
  //@}
};

/**
 * Subclass of SocketDescriptor to provide TCP client-socket functionality. */
class ClientSocketDescriptor: public SocketDescriptor {
public:
  /// Constructs a TCP socket connected to the given hostname/port combination.
  ClientSocketDescriptor(ref<FileDescriptorManager> const &_fdMgr,
			 string const &_hostname, int _port);
};

/**
 * Subclass of SocketDescriptor to provide TCP server-socket functionality. */
class ServerSocketDescriptor: public SocketDescriptor {
private:
  string canonicalAddress;	/// Local end's xxx.xxx.xxx.xxx:yyyyy IP addr/port number

public:
  /**
   * Construct a TCP server-socket listening on the given port, unless
   * 0 is specified, in which case an unused port is used, on the
   * given interface-name, unless the empty-string is specified, in
   * which case INADDR_ANY is used (ie. bind() to all interfaces if no
   * preference is supplied). */
  ServerSocketDescriptor(ref<FileDescriptorManager> const &_fdMgr,
			 int _port = 0, string const &_hostname = "");

  /**
   * Calls ::accept() and returns a SocketDescriptor for the resulting
   * new connection. Throws FileDescriptor::Exception on failure. */
  ref<SocketDescriptor> accept();

  /// Retrieves the IP address and port number that this server socket is bound to.
  string const &getCanonicalAddress() const { return canonicalAddress; }
};

///////////////////////////////////////////////////////////////////////////
// Utility classes

template <class ConnectionHandler> class SocketAcceptorCallback: public FileDescriptor::Callback {
private:
  void acceptHandler(ref<FileDescriptor> const &desc) {
    ref<ServerSocketDescriptor> server = (ServerSocketDescriptor *) desc.get();
    ref<Referenceable> dummyRef = new ConnectionHandler(server->accept());
  }

public:
  SocketAcceptorCallback(ref<ServerSocketDescriptor> const &server) {
    server->setReadCallback(this,
			    (FileDescriptor::Callback::Method)
			      &SocketAcceptorCallback::acceptHandler);
  }
};

#endif
