/**************************************************************************
 Copyright (c) 2000-2001, Tony Garnock-Jones
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:

     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above
       copyright notice, this list of conditions and the following
       disclaimer in the documentation and/or other materials provided
       with the distribution.

     * Neither the names of the copyright holders nor the names of this
       software's contributors may be used to endorse or promote
       products derived from this software without specific prior
       written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

/* -*- c++ -*- */
#ifndef MPS_Connection_H
#define MPS_Connection_H

namespace MPS {

/**
 * Manages a connection over a Transport to an MPS server. Allows
 * message reception and transmission. */
class Connection: public Referenceable {
public:
  /// A callback - subclass, and use with setCallback.
  class Callback: public Referenceable {
  public:
    typedef void (Callback::*Method)(InputStream &input);
  };

private:
  Address address;		///< the address of the remote object
  ref<Callback> callback;		///< currently-registered callback.
  Callback::Method callbackMethod;	///< method to invoke on registered callback.

protected:
  /// Constructor for subclasses.
  Connection(Address const &a)
    : address(a),
      callback(0),
      callbackMethod(0)
  {}

public:
  virtual ~Connection() {}

  /// Extract the address of the object we're connected to.
  Address const &getRemoteAddress() const { return address; }

  /**
   * Obtain an OutputStream to use to send a message to the remote
   * object. The caller must call releaseOutputStream() when it's
   * finished with the stream.
   *
   * @return an OutputStream which, when flushed, will send a message
   * to the remote object. */
  virtual OutputStream *getOutputStream() = 0;

  /**
   * Signals that the caller is finished with an OutputStream obtained
   * via the getOutputStream() method.
   *
   * @param stream the stream to release */
  virtual void releaseOutputStream(OutputStream *stream) = 0;

  /**
   * Obtain an InputStream to use to receive a message from the remote
   * object. The caller must call releaseInputStream() when it's
   * finished with the stream.
   *
   * @return an InputStream which allows reading of the most recent
   * message sent back from the remote object. */
  virtual InputStream *getInputStream() = 0;

  /**
   * Signals that the caller is finished with an InputStream obtained
   * via the getInputStream() method.
   *
   * @param stream the stream to release */
  virtual void releaseInputStream(InputStream *stream) = 0;

  /**
   * Registers a callback.  Override this if you need to do special
   * processing on callback registration (unlikely).
   *
   * @see Proxy::_setCallback */
  virtual void setCallback(ref<Callback> _callback, Callback::Method _method) {
    callback = _callback;
    callbackMethod = _method;
  }

  /**
   * If a callback is registered, deregisters it, and calls it. Note
   * that callbacks should be called when it's safe to read a complete
   * message from this connection (ie. it won't block to do so). This
   * method should be called from somewhere within Transports' main loops.
   *
   * @see Proxy::_setCallback */
  void fireCallback(InputStream &input) {
    if (callback.valid()) {
      // Save the existing callback...
      ref<Callback> cb = callback;
      Callback::Method meth = callbackMethod;

      // Clear it out, so there's no registered callback anymore...
      callback = 0;
      callbackMethod = 0;

      // ... and finally call our saved callback.
      (cb->*meth)(input);
    } else {
      // Do nothing if no callback is registered.
    }
  }
};

}

#endif
