/* -*- c++ -*- */
/**************************************************************************
 Copyright (c) 2000-2001, Tony Garnock-Jones
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:

     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above
       copyright notice, this list of conditions and the following
       disclaimer in the documentation and/or other materials provided
       with the distribution.

     * Neither the names of the copyright holders nor the names of this
       software's contributors may be used to endorse or promote
       products derived from this software without specific prior
       written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

#ifndef MPS_Transport_H
#define MPS_Transport_H

namespace MPS {

/**
 * Represents a method of resolving names for, and connecting to
 * remote objects. Subclasses implement name-resolution and connection
 * facilities according to a particular transport protocol; examples
 * include InetTransport, SleeTransport, and FastmsgTransport. */
class Transport {
private:
  typedef vector<Transport *> transportVec_t;
  static transportVec_t allTransports;			///< All registered Transports

  typedef map<string, ref<Interface> > ifMap_t;
  static ifMap_t localInterfaces;			///< All locally-registered interfaces

private:
  string name;				///< Registered name of this transport ("inet", "slee", ..)

private:
  /**
   * Locate a Transport by name.
   *
   * @return a pointer to the named Transport.
   * @exception MPSException if the Transport name is not registered. */
  static Transport *findTransport(string const &name) {
    for (transportVec_t::iterator i = allTransports.begin();
	 i != allTransports.end();
	 i++) {
      if ((*i)->name == name)
	return (*i);
    }

    throw MPSException("Transport " + name + " not registered");
  }

  /**
   * Bind an interface in the specially-implemented "local" transport.
   */
  static void bindLocalInterface(InterfaceServer *server);

public:
  /// Register a transport.
  static void registerTransport(Transport *t) {
    allTransports.push_back(t);
  }

  /**
   * Attempts to resolve an object name within all known address
   * spaces, in order of registration with registerTransport().
   *
   * @param objectName the unresolved name of the object we're looking for
   * @return the empty string if the object name does not exist or the
   * resolved name of the object if it does. */
  static string resolveName(string const &objectName, bool includeLocal = false) {
    if (objectName.substr(0, 4) == "mps:")
      return objectName;

    if (includeLocal) {
      ifMap_t::iterator ii = localInterfaces.find(objectName);
      if (ii != localInterfaces.end()) {
	char buf[64];
	sprintf(buf, "mps:local:%lu", (long) (*ii).second.get());
	return buf;
      }
    }

    for (transportVec_t::iterator i = allTransports.begin();
	 i != allTransports.end();
	 i++) {
      string result = (*i)->resolveObjectName(objectName);
      if (result.size() > 0)
	return result;
    }

    return "";
  }

  /**
   * Returns a message source/sink that can be used for communication
   * with the object at the address described by the
   * connectionSpec. Selects an appropriate Transport based on the
   * "transport" field of the Address.
   *
   * @param connectionSpec the address of the remote object
   * @return a connection handle for remote communication, or NULL if
   * none could be created
   * @exception MPSException if the transport name given is not
   * registered */
  static ref<Connection> getConnectionTo(Address const &connectionSpec) {
    return findTransport(connectionSpec.getTransport())->connectTo(connectionSpec);
  }

  /**
   * Bind an object server to an interface provided by a transport.
   *
   * @param server the server to bind to the (about-to-be-created) transport-interface
   *
   * @param resolvedName the suggested resolved name to bind to -
   * should be of the form "mps:transportname" or, if a greater degree
   * of control over the bind is desired,
   * "mps:transportname:param:param:..." (for instance, with the
   * "inet" transport, you use "mps:inet:hostname:portnumber"). */
  static string bindServer(Server *server, string const &resolvedName) {
    Address spec(resolvedName);
    return findTransport(spec.getTransport())->registerServer(server, spec);
  }

  /**
   * Special-case binding of local interface-providing servers, so
   * that the "local" transport will work for them. Note that
   * non-interface-providing servers can therefore not operate over
   * the "local" transport.
   *
   * @see bindServer(Server *, string const &)
   */
  static string bindServer(InterfaceServer *server, string const &resolvedName) {
    bindLocalInterface(server);
    string canonicalName = bindServer((Server *) server, resolvedName);
    server->__impl->_setResolvedName(canonicalName);
    return canonicalName;
  }

  /**
   * Bind an object server to all available transports that currently
   * do not have a binding for this server.
   *
   * @param server the server to bind */
  static string bindServer(Server *server) {
    string canonicalName;

    for (transportVec_t::iterator i = allTransports.begin();
	 i != allTransports.end();
	 i++) {
      Transport *t(*i);
      Address spec("mps:" + t->getName());
      canonicalName = t->registerServer(server, spec); // keep the last one. This is a hack :-(
    }

    return canonicalName;
  }

  /**
   * Special-case binding of local interface-providing servers, so
   * that the "local" transport will work for them. Note that
   * non-interface-providing servers can therefore not operate over
   * the "local" transport.
   *
   * @see bindServer(Server *)
   */
  static string bindServer(InterfaceServer *server) {
    bindLocalInterface(server);
    string canonicalName = bindServer((Server *) server);
    if (canonicalName != "")
      server->__impl->_setResolvedName(canonicalName);
    return canonicalName;
  }

protected:
  /**
   * For access by subclasses - initialises this class
   *
   * @param _name name to register the transport under
   */
  Transport(string const &_name)
    : name(_name)
  {
    registerTransport(this);
  }

public:
  virtual ~Transport() {}

  /// Retrieves the transport's registered name.
  string const &getName() const { return name; }

protected:
  /**
   * Returns a message source/sink that can be used for communication
   * with the object at the address described by the
   * connectionSpec. The connectionSpec is the broken-down resolved
   * name of the remote object, and <b>must</b> match the address
   * format that this transport expects.
   *
   * @param connectionSpec the address of the remote object
   * @return a connection handle for remote communication, or NULL if
   * none could be created */
  virtual ref<Connection> connectTo(Address const &connectionSpec) = 0;

  /**
   * Attempts to resolve an object name within this transport's known
   * object space.
   *
   * @param objectName the unresolved name of the object we're looking for
   * @return the empty string if the object does not exist in this transport's
   * sphere of knowledge, or the resolved name of the object if it does.
   */
  virtual string resolveObjectName(string const &objectName) const = 0;

  /**
   * Attempt to register a server interface for access using this transport.
   *
   * @param server the server to register
   * @param spec the connection specification that the transport may
   * use as a hint in determining the final resolved name to use when
   * binding this object.
   * @return the canonical name of the server-bound-to-this-transport,
   * or "" if binding to this transport fails. */
  virtual string registerServer(Server *server, Address const &spec) = 0;
};

}

#endif
