/* -*- c++ -*- */
/**************************************************************************
 Copyright (c) 2000-2001, Tony Garnock-Jones
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:

     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above
       copyright notice, this list of conditions and the following
       disclaimer in the documentation and/or other materials provided
       with the distribution.

     * Neither the names of the copyright holders nor the names of this
       software's contributors may be used to endorse or promote
       products derived from this software without specific prior
       written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

#ifndef MPS_Transport_Inet_H
#define MPS_Transport_Inet_H

#include <mps/mps.h>
#include <FileDescriptor.h>
#include <SocketDescriptor.h>

namespace MPS {

/**
 * Implements a TCP/IP-socket Transport class for accessing and
 * serving MPS objects. Manages all server and client sockets for this
 * process. Dispatches incoming work to the appropriate objects. */
class InetTransport: public MPS::Transport {
private:
  static InetTransport *instance;	///< The single, per-application instance of InetTransport

  //---------------------------------------------------------------------------

  class InetIOStream;		///< Subclass of MPS::InputStream and MPS::OutputStream
  class InetConnection;		///< Subclass of MPS::Connection - holds demux, oid
  class InetDemux;		///< Reads OID from an incoming message and dispatches to callback
  class InetDispatcher;		///< Callback for dispatching to a given MPS::Server

  typedef map< string, ref<InetDemux> > demuxMap_t;	///< Map of inet addr -> demux

  //---------------------------------------------------------------------------

  ref<FileDescriptorManager> fdMgr;	///< All FileDescriptors we are responsible for.

  demuxMap_t clientMap;			///< Find a client demux for a given ipaddr:port
  demuxMap_t serverMap;			///< Find a server demux for a given ipaddr:port
  ref<InetDemux> defaultDemux;		///< Demultiplexor for the server socket
  int nextOid;				///< Next OID available for registration here.

  string namingServerLocation;		///< Resolved name of the naming server to use
  ref<NamingService> namingServer;	///< Reference to the naming server object

  //---------------------------------------------------------------------------

  InetTransport(string const &nsLoc,
		string const &hostName,
		int portNumber);

  ref<InetDemux> registerServerSocket(ref<ServerSocketDescriptor> const &serverSock);

protected:
  virtual ref<Connection> connectTo(Address const &connectionSpec);
  virtual string resolveObjectName(string const &objectName) const;
  virtual string registerServer(Server *server, Address const &spec);

  ref<FileDescriptor> socketFor(Address const &connectionSpec);

public:
  /**
   * Creates the single per-application instance of InetTransport, and
   * registers it using Transport::registerTransport().
   *
   * @param namingServerLocation resolved address of the naming server to use
   * @param hostName either "" (for all) or the name of the local interface to bind to
   * @param portNumber either 0 (for any) or the number of the TCP port to bind to */
  static void initialise(string const &namingServerLocation,
			 string const &hostName,
			 int portNumber) {
    if (instance == 0) {
      instance = new InetTransport(namingServerLocation,
				   hostName,
				   portNumber);
    }
  }

  /**
   * Once you have access to the FileDescriptorManager, you can call
   * the mainloop(), processFdSet() or poll() methods on it to service
   * requests managed by this Transport. */
  static ref<FileDescriptorManager> const &getFdMgr() {
    return instance->fdMgr;
  }
};

}

#endif
